/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2009-2010 Canonical Services Ltd (www.canonical.com)
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <string.h>
#include <libxml/HTMLparser.h>
#include <webkit/webkit.h>
#include <dbus/dbus-glib.h>
#include <gnome-keyring.h>
#include "oauth.h"
#include "u1-music-store.h"
#include "syncdaemon.h"
#include "u1-marshal.h"

#define U1_MUSIC_LIBRARY_LOCATION  ".ubuntuone/Purchased from Ubuntu One"

#define U1_STORE_URL               "/music/login"
#define U1_LIBRARY_URL             "/music/store/library"
#define U1_NOT_LOGGED_IN_STORE_URL "/music/store-no-token"
#define U1_NOT_REGISTERED_URL      "https://one.ubuntu.com/music/notregistered?returnUrl="

#define U1_INITIAL_PAGE "<html><body>Loading Ubuntu One music store</body></html>"
#define U1_DEFAULT_ERROR_PAGE "<html><body>Could not load Music Store</body></html>"
#define U1_IN_DEVELOPMENT_PAGE "<html><body>The music store is under development, please be patient</body></html>"
#define U1_CONNECTING_PAGE "<html><body>Connecting to Ubuntu One. Please wait...</body></html>"

struct _U1MusicStorePrivate {
	DBusGConnection *bus;
	gchar *base_url;

	GtkWidget *scroll;
	GtkWidget *initial_web_viewer;
	GtkWidget *web_viewer;
	WebKitWebFrame *frame_to_use;

	GdkCursor *busy_cursor;
	GdkCursor *normal_cursor;

	guint watch_id;
	GHashTable *watched_downloads;
};

G_DEFINE_TYPE(U1MusicStore, u1_music_store, GTK_TYPE_VBOX)

enum {
	PREVIEW_MP3_SIGNAL,
	PLAY_LIBRARY_SIGNAL,
	URL_LOADED_SIGNAL,
	DOWNLOAD_FINISHED_SIGNAL,
	LAST_SIGNAL
};

static guint u1_music_store_signals[LAST_SIGNAL] = { 0, };

static void
u1_music_store_finalize (GObject *object)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (object);

	syncdaemon_finalize ();

	if (music_store->priv != NULL) {
		if (music_store->priv->watch_id != 0) {
			g_source_remove (music_store->priv->watch_id);
			music_store->priv->watch_id = 0;
		}

		if (music_store->priv->watched_downloads != NULL) {
			g_hash_table_destroy (music_store->priv->watched_downloads);
			music_store->priv->watched_downloads = NULL;
		}

		if (music_store->priv->base_url != NULL) {
			g_free (music_store->priv->base_url);
			music_store->priv->base_url;
		}

		if (music_store->priv->bus != NULL) {
			dbus_g_connection_unref (music_store->priv->bus);
			music_store->priv->bus = NULL;
		}

		g_free (music_store->priv);
		music_store->priv = NULL;
	}

	G_OBJECT_CLASS (u1_music_store_parent_class)->finalize (object);
}

static void
u1_music_store_class_init (U1MusicStoreClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	/* Signals */
	u1_music_store_signals[PREVIEW_MP3_SIGNAL] = g_signal_new ("preview-mp3",
								   G_TYPE_FROM_CLASS (klass),
								   (GSignalFlags) G_SIGNAL_RUN_LAST,
								   G_STRUCT_OFFSET (U1MusicStoreClass, preview_mp3),
								   NULL,
								   NULL,
								   _u1_marshal_VOID__STRING_STRING,
								   G_TYPE_NONE, 2,
								   G_TYPE_STRING, G_TYPE_STRING);
	u1_music_store_signals[PLAY_LIBRARY_SIGNAL] = g_signal_new ("play-library",
								    G_TYPE_FROM_CLASS (klass),
								    (GSignalFlags) G_SIGNAL_RUN_LAST,
								    G_STRUCT_OFFSET (U1MusicStoreClass, play_library),
								    NULL,
								    NULL,
								    g_cclosure_marshal_VOID__STRING,
								    G_TYPE_NONE, 1,
								    G_TYPE_STRING);
	u1_music_store_signals[URL_LOADED_SIGNAL] = g_signal_new ("url-loaded",
								  G_TYPE_FROM_CLASS (klass),
								  (GSignalFlags) G_SIGNAL_RUN_LAST,
								  G_STRUCT_OFFSET (U1MusicStoreClass, url_loaded),
								  NULL,
								  NULL,
								  g_cclosure_marshal_VOID__STRING,
								  G_TYPE_NONE, 1,
								  G_TYPE_STRING);
	u1_music_store_signals[DOWNLOAD_FINISHED_SIGNAL] = g_signal_new ("download-finished",
									 G_TYPE_FROM_CLASS (klass),
									 (GSignalFlags) G_SIGNAL_RUN_LAST,
									 G_STRUCT_OFFSET (U1MusicStoreClass, download_finished),
									 NULL,
									 NULL,
									 g_cclosure_marshal_VOID__STRING,
									 G_TYPE_NONE, 1,
									 G_TYPE_STRING);

	object_class->finalize = u1_music_store_finalize;
}

static void
parse_oauth_string (const gchar *string, gchar **oauth_token, gchar **oauth_token_secret)
{
	gchar **items;

	*oauth_token = NULL;
	*oauth_token_secret = NULL;

	items = g_strsplit (string, "&", 2);
	if (items) {
		gint i;

		for (i = 0; items[i] != NULL; i++) {
			gchar **kv_pair;

			kv_pair = g_strsplit (items[i], "=", 2);
			if (kv_pair) {
				if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_token") == 0)
					*oauth_token = g_strdup (kv_pair[1]);
				else if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_token_secret") == 0)
					*oauth_token_secret = g_strdup (kv_pair[1]);

				g_strfreev (kv_pair);
			}
		}

		g_strfreev (items);
	}
}

static void
get_credentials_from_keyring (gchar **oauth_token, gchar **oauth_token_secret)
{
	GnomeKeyringAttributeList *attrs;
	GnomeKeyringResult result;
	GList *items_found;

	*oauth_token = *oauth_token_secret = NULL;

	/* Get the OAuth token from the keyring */
	attrs = gnome_keyring_attribute_list_new ();
	gnome_keyring_attribute_list_append_string (attrs, "ubuntuone-realm", "https://ubuntuone.com");
	gnome_keyring_attribute_list_append_string (attrs, "oauth-consumer-key", "ubuntuone");

	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
						attrs, &items_found);
	if (result == GNOME_KEYRING_RESULT_OK && items_found != NULL) {
		GnomeKeyringFound *first_item = (GnomeKeyringFound *) items_found->data;
					
		parse_oauth_string (first_item->secret, oauth_token, oauth_token_secret);
		gnome_keyring_found_list_free (items_found);

		/* Connect to syncdaemon now that we have tokens */
		syncdaemon_connect ();
	}
}

typedef struct {
	U1MusicStore *music_store;
	gchar *not_registered_url;
} NotRegisteredReplacementData;

static void
got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar *consumer_key, gpointer user_data)
{
	gchar *oauth_token, *oauth_token_secret, *url_to_use, *real_url;
	gchar **uri_parts;
	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;

	get_credentials_from_keyring (&oauth_token, &oauth_token_secret);

	/* Get the return_url parameter from the not-registered url */
	uri_parts = g_strsplit (nrrd->not_registered_url, "?", 0);
	if (uri_parts != NULL && uri_parts[1] != NULL) {
		gchar **uri_args;

		uri_args = g_strsplit (uri_parts[1], "?", 0);
		if (uri_args != NULL) {
			gint i;

			for (i = 0; uri_args[i] != NULL; i++) {
				if (g_str_has_prefix (uri_args[i], "returnUrl")) {
					gchar **this_arg;

					this_arg = g_strsplit (uri_args[i], "=", 2);
					url_to_use = g_strdup_printf ("%s%s?returnUrl=%s",
								      nrrd->music_store->priv->base_url,
								      U1_STORE_URL,
								      this_arg[1]);

					g_strfreev (this_arg);
					break;
				}
			}

			g_strfreev (uri_args);
		}

		g_strfreev (uri_parts);
	} else
		url_to_use = g_strdup_printf ("%s%s", nrrd->music_store->priv->base_url, U1_STORE_URL);

	/* Now sign the URL and load the new page in the view */
	real_url = oauth_sign_url2 (url_to_use, NULL, OA_HMAC, "GET",
				    "ubuntuone", "hammertime",
				    oauth_token, oauth_token_secret);
	webkit_web_view_open (WEBKIT_WEB_VIEW (nrrd->music_store->priv->web_viewer), real_url);

	/* Free memory */
	g_object_unref (G_OBJECT (proxy));
	g_free (url_to_use);
	g_free (real_url);
	g_free (oauth_token);
	g_free (oauth_token_secret);

	g_free (nrrd->not_registered_url);
	g_free (nrrd);
}

static void
authorization_denied_cb (DBusGProxy *proxy, gpointer user_data)
{
	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;

	/* Free memory */
	g_free (nrrd->not_registered_url);
	g_free (nrrd);
}

static void
oauth_error_cb (DBusGProxy *proxy, const gchar *message, gpointer user_data)
{
	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;

	g_debug ("OAuth error: %s", message);

	/* Free memory */
	g_free (nrrd->not_registered_url);
	g_free (nrrd);
}

static gboolean
navigation_requested_cb (WebKitWebView *web_view,
			 WebKitWebFrame *frame,
			 WebKitNetworkRequest *request,
			 WebKitWebNavigationAction *navigation_action,
			 WebKitWebPolicyDecision *policy_decision,
			 gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	/* Remove watching callback for the page */
	if (music_store->priv->watch_id != 0) {
		g_source_remove (music_store->priv->watch_id);
		music_store->priv->watch_id = 0;
	}

	if (music_store->priv->watched_downloads != NULL) {
		g_hash_table_destroy (music_store->priv->watched_downloads);
		music_store->priv->watched_downloads = NULL;
	}

	/* If the URL just loaded is for not registered users, fire the OAuth authentication */
	if (g_str_has_prefix (webkit_network_request_get_uri (request), U1_NOT_REGISTERED_URL)) {
		DBusGProxy *proxy;
		gboolean success;
		GError *error = NULL;

		webkit_web_view_load_string (web_view, U1_CONNECTING_PAGE, "text/html", "utf-8", "file:///");

		proxy = dbus_g_proxy_new_for_name (music_store->priv->bus, "com.ubuntuone.Authentication",
						   "/", "com.ubuntuone.Authentication");
		if (proxy != NULL) {
			NotRegisteredReplacementData *nrrd;

			nrrd = g_new0 (NotRegisteredReplacementData, 1);
			nrrd->music_store = music_store;
			nrrd->not_registered_url = g_strdup (webkit_network_request_get_uri (request));

			dbus_g_proxy_add_signal (proxy, "AuthorizationDenied", G_TYPE_INVALID);
			dbus_g_proxy_connect_signal (proxy, "AuthorizationDenied",
						     G_CALLBACK (authorization_denied_cb),
						     nrrd,
						     NULL);

			dbus_g_proxy_add_signal (proxy, "NoCredentials", G_TYPE_INVALID);
			dbus_g_proxy_connect_signal (proxy, "NoCredentials",
						     G_CALLBACK (authorization_denied_cb),
						     nrrd,
						     NULL);
			
			dbus_g_proxy_add_signal (proxy, "NewCredentials", G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
			dbus_g_proxy_connect_signal (proxy, "NewCredentials",
						     G_CALLBACK (got_new_dbus_credentials_cb),
						     nrrd,
						     NULL);

			dbus_g_proxy_add_signal (proxy, "OAuthError", G_TYPE_STRING, G_TYPE_INVALID);
			dbus_g_proxy_connect_signal (proxy, "OAuthError",
						     G_CALLBACK (oauth_error_cb),
						     nrrd, NULL);

			error = NULL;
			success = dbus_g_proxy_call (proxy, "login", &error,
						     G_TYPE_STRING, "https://ubuntuone.com",
						     G_TYPE_STRING, "ubuntuone",
						     G_TYPE_INVALID,
						     G_TYPE_INVALID);
			if (!success) {
				g_warning ("DBus call error: %s", error->message);
				g_error_free (error);
			}
		} else {
			g_warning ("Couldn't get proxy for com.ubuntuone.Authentication");
		}

		return FALSE;
	} else {
		const gchar *library_override;

		library_override = g_getenv ("U1MUSICLIBRARYURL");

		/* If the URL being loaded is the 7d library page, and we have a library
		   override set, load ours instead. */
		if (g_strrstr (webkit_network_request_get_uri (request), "library.aspx") != NULL ||
		    g_str_has_prefix (webkit_network_request_get_uri (request), "https://one.ubuntu.com/music/store/library")) {
			if (library_override != NULL) {
				g_debug ("overriding the library page to ours");
				webkit_web_view_open (web_view, library_override);
			}
		}
	}

	/* Change the cursor when loading */
	if (music_store->priv->normal_cursor)
		gdk_cursor_unref (music_store->priv->normal_cursor);
#if GTK_CHECK_VERSION(2, 18, 0)
	music_store->priv->normal_cursor = gdk_window_get_cursor (GTK_WIDGET (web_view)->window);
#endif
	gdk_cursor_ref (music_store->priv->normal_cursor);

	if (music_store->priv->busy_cursor == NULL)
		music_store->priv->busy_cursor = gdk_cursor_new_for_display (gdk_display_get_default (), GDK_WATCH);

	gdk_window_set_cursor (GTK_WIDGET (web_view)->window, music_store->priv->busy_cursor);

	/* Inform webkit to continue with the loading */
	webkit_web_policy_decision_use (policy_decision);

	return TRUE;
}

static void
execute_script (WebKitWebView *web_view, const gchar *script_name, const gchar *replacement)
{
	gchar *file_contents, *script, *path;
	gsize length;
	GError *error = NULL;

	path = g_build_path ("/", U1_JAVASCRIPT_DIR, script_name, NULL);

	/* Load script from known location */
	if (g_file_get_contents (path, &file_contents, &length, &error)) {
		if (replacement != NULL)
			script = g_strdup_printf (file_contents, replacement);
		else
			script = g_strdup (file_contents);

		webkit_web_view_execute_script (web_view, script);

		g_free (script);
		g_free (file_contents);
	} else {
		g_warning ("Could not load %s: %s", path, error->message);
		g_error_free (error);
	}

	g_free (path);
}

static void
parse_html_node (U1MusicStore *music_store, WebKitWebView *web_view, htmlNodePtr node, GHashTable *current_downloads)
{
	htmlNodePtr subnode;
	GList *keys;
	static GHashTable *downloads_in_progress = NULL;

	if (node == NULL)
		return;

	/* We keep track of all the files being downloaded */
	if (downloads_in_progress == NULL)
		downloads_in_progress = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

	keys = g_hash_table_get_keys (current_downloads);
	while (keys != NULL) {
		g_hash_table_insert (downloads_in_progress, g_strdup (keys->data), g_strdup (keys->data));
		keys = g_list_remove (keys, keys->data);
	}

	if (music_store->priv->watched_downloads == NULL)
		music_store->priv->watched_downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

	/* Search HTML for downloads */
	for (subnode = node->children; subnode != NULL; subnode = subnode->next) {
		if (!g_strcmp0 ((const gchar *) subnode->name, "li")) {
			xmlChar *prop;
			gchar *script = NULL, *full_path;
			GString *escaped_prop;
			gint i;
			SyncdaemonDownload *download;

			prop = xmlGetProp (subnode, "data-path");
			if (prop == NULL)
				continue;

			/* Escape 's in the file name */
			escaped_prop = g_string_new (NULL);
			for (i = 0; i < strlen (prop); i++) {
				if (prop[i] == '\'')
					escaped_prop = g_string_append (escaped_prop, "\\\'");
				else
					escaped_prop = g_string_append_c (escaped_prop, prop[i]);
			}

			full_path = g_build_filename (g_get_home_dir (),
						      U1_MUSIC_LIBRARY_LOCATION,
						      prop, NULL);

			download = g_hash_table_lookup (current_downloads, full_path);
			if (download != NULL) {
				gdouble percent;

				if (download->deflated_size != 0)
					percent = (gdouble) download->n_bytes_read / (gdouble) download->deflated_size;
				else
					percent = 0.0;
				script = g_strdup_printf ("setProgressBar('%s', %.2f, '%.1fMB of %.1fMB read');",
							  escaped_prop->str, percent,
							  (gdouble) download->n_bytes_read / (gdouble) 1024 / (gdouble) 1024,
							  (gdouble) download->deflated_size / (gdouble) 1024 / (gdouble) 1024);
			} else {
				/* If the file exists, the download is done */
				if (g_file_test (full_path, G_FILE_TEST_EXISTS)) {
					if (!g_hash_table_lookup (music_store->priv->watched_downloads, (gconstpointer) full_path)) {
						if (g_hash_table_lookup (downloads_in_progress, full_path)) {
							script = g_strdup_printf ("setProgressBar('%s', 1.0, 'Completed', false);",
										  escaped_prop->str);
							g_hash_table_remove (downloads_in_progress, full_path);
						} else {
							script = g_strdup_printf ("setProgressBar('%s', 1.0, 'Completed', true);",
										  escaped_prop->str);
						}

						g_signal_emit (music_store, u1_music_store_signals[DOWNLOAD_FINISHED_SIGNAL], 0,
							       (const gchar *) full_path);

						/* Add the file to watched downloads so that we don't signal again */
						g_hash_table_insert (music_store->priv->watched_downloads,
								     g_strdup (full_path),
								     g_strdup (full_path));
					}
				} else
					script = g_strdup_printf ("setProgressBar('%s', 0.0, 'Transferring to your Ubuntu One storage');",
								  escaped_prop->str);
			}

			if (script != NULL) {
				webkit_web_view_execute_script (web_view, script);
				g_free (script);
			}

			g_free (full_path);
			g_string_free (escaped_prop, TRUE);
		} else
			parse_html_node (music_store, web_view, subnode, current_downloads);
	}
}

static void
update_status_in_library_page (U1MusicStore *music_store, WebKitWebView *web_view, WebKitWebFrame *frame)
{
	GString *html;
	WebKitWebDataSource *data_source;
	htmlDocPtr html_doc;
	htmlNodePtr root_node;
	GHashTable *current_downloads;

	/* Retrieve contents of the page */
	data_source = webkit_web_frame_get_data_source (frame);
	html = webkit_web_data_source_get_data (data_source);

	/* Get list of current downloads */
	current_downloads = syncdaemon_get_current_downloads ();
	if (current_downloads != NULL) {
		/* Parse HTML */
		html_doc = htmlParseDoc (html->str, NULL);
		root_node = xmlDocGetRootElement (html_doc);
		parse_html_node (music_store, web_view, root_node, current_downloads);

		/* Free memory */
		xmlFreeDoc (html_doc);
		g_hash_table_destroy (current_downloads);
	}
}

static gboolean
poll_downloads_cb (gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	update_status_in_library_page (music_store,
				       WEBKIT_WEB_VIEW (music_store->priv->web_viewer),
				       music_store->priv->frame_to_use);

	return TRUE;
}

static void
load_finished_cb (WebKitWebView *web_view, WebKitWebFrame *frame, gpointer user_data)
{
	const gchar *library_override;
	gchar *tmp_url;
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	/* Remove the initial view if still around */
	if (music_store->priv->initial_web_viewer != NULL) {
		gtk_widget_destroy (music_store->priv->initial_web_viewer);
		music_store->priv->initial_web_viewer = NULL;

		/* And show the real store viewer */
		g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "navigation-policy-decision-requested",
				  G_CALLBACK (navigation_requested_cb), music_store);
		gtk_widget_show (music_store->priv->web_viewer);
		gtk_container_add (GTK_CONTAINER (music_store->priv->scroll), music_store->priv->web_viewer);
	}

	/* Inject some JavaScript to enable previews and hide the Flash player */
	execute_script (web_view, "u1-preview.js", NULL);

	/* Allow us to specifically override the library page links */
	library_override = g_getenv ("U1MUSICLIBRARYURL");
	if (library_override != NULL)
		execute_script (web_view, "u1-library-override.js", library_override);

	gdk_window_set_cursor (GTK_WIDGET (web_view)->window, music_store->priv->normal_cursor);

	/* If the URL just loaded is the library page, update download status */
	tmp_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_LIBRARY_URL);
	if (g_str_has_prefix (webkit_web_view_get_uri (web_view), tmp_url)) {
		/* Inject JavaScript to make songs clickable 
		   Needs to be done before update-status-in-library-page */
		execute_script (web_view, "u1-songs-clickable.js", NULL);

		music_store->priv->frame_to_use = frame;
		update_status_in_library_page (music_store, web_view, frame);

		/* Poll syncdaemon for downloads while the page is loaded */
		music_store->priv->watch_id = g_timeout_add (5000, (GSourceFunc) poll_downloads_cb, music_store);
	}

	g_free (tmp_url);

	/* Signal users the URL finished loading */
	g_signal_emit (music_store, u1_music_store_signals[URL_LOADED_SIGNAL], 0,
		       (const gchar *) webkit_web_view_get_uri (WEBKIT_WEB_VIEW (music_store->priv->web_viewer)));
}

static void
status_bar_text_changed_cb (WebKitWebView *web_view, const gchar *text, gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	if (g_str_has_prefix (text, "u1preview:::")) {
		gchar **args;

		args = g_strsplit (text, ":::", 0);
		if (args != NULL) {
			g_signal_emit (music_store, u1_music_store_signals[PREVIEW_MP3_SIGNAL], 0, args[1], args[2]);
			g_strfreev (args);
		}
	}
	if (g_str_has_prefix (text, "u1playlibrary:::")) {
		gchar **args;

		args = g_strsplit (text, ":::", 2);
		if (args != NULL) {
			g_signal_emit (music_store, u1_music_store_signals[PLAY_LIBRARY_SIGNAL], 0, args[1]);
			g_strfreev (args);
		}
	}
}

static gboolean
load_error_cb (WebKitWebView *web_view, WebKitWebFrame *frame, const gchar *uri, GError *error, gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	webkit_web_view_load_string (web_view, U1_DEFAULT_ERROR_PAGE, "text/html", "utf-8", "file:///");

	return TRUE;
}

static gchar *
get_url_to_use (U1MusicStore *music_store)
{
	gchar *oauth_token = NULL, *oauth_token_secret = NULL, *real_url;
	const gchar *url_to_use;

	/* If U1MUSICSTOREURL is defined, use that instead of the real URL */
	if (! (url_to_use = g_getenv ("U1MUSICSTOREURL")))
		url_to_use = "https://one.ubuntu.com";

	if (music_store->priv->base_url != NULL)
		g_free (music_store->priv->base_url);

	music_store->priv->base_url = g_strdup (url_to_use);

	/* OAUth sign the URL */
	if (g_str_has_prefix (url_to_use, "http://localhost") || g_str_has_prefix (url_to_use, "http://127.0.0.1")) {
		const gchar *oauthfile;

		oauthfile = g_getenv ("OAUTHKEYFILE");
		if (oauthfile != NULL) {
			gchar *oauthstring;
			gsize len;

			if (g_file_get_contents (oauthfile, &oauthstring, &len, NULL)) {
				parse_oauth_string (oauthstring, &oauth_token, &oauth_token_secret);
				g_free (oauthstring);
			}
		} else {
			webkit_web_view_load_string (WEBKIT_WEB_VIEW (music_store->priv->web_viewer),
						     "Fail! If you specify U1MUSICSTOREURL you must "
						     "also specify OAUTHKEYFILE as a file with OAuth keys in it",
						     "text/html", "utf-8", "file:///");
			return NULL;
		}
	} else
		get_credentials_from_keyring (&oauth_token, &oauth_token_secret);

	if (oauth_token == NULL || oauth_token_secret == NULL)
		real_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_NOT_LOGGED_IN_STORE_URL);
	else {
		gchar *s = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_STORE_URL);

		real_url = oauth_sign_url2 (s, NULL, OA_HMAC, "GET",
					    "ubuntuone", "hammertime",
					    oauth_token, oauth_token_secret);
		g_free (s);
	}

	g_free (oauth_token);
	g_free (oauth_token_secret);

	return real_url;
}

static gboolean
load_real_store_cb (gpointer user_data)
{
	gchar *real_url;
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	g_debug ("Loading the real store page");
	real_url = get_url_to_use (music_store);
	if (real_url != NULL) {
		webkit_web_view_open (WEBKIT_WEB_VIEW (music_store->priv->web_viewer), real_url);

		/* Free memory */
		g_free (real_url);
	}

	return FALSE;
}

static void
u1_music_store_init (U1MusicStore *music_store)
{
	gchar *new_user_agent;
	GError *error = NULL;

	if (!g_thread_get_initialized ())
		g_thread_init (NULL);

	music_store->priv = g_new0 (U1MusicStorePrivate, 1);
	music_store->priv->watched_downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

	/* Initialize DBus */
	music_store->priv->bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (error) {
		g_warning ("Couldn't get session bus: %s", error->message);
		g_error_free (error);

	}

	dbus_g_object_register_marshaller (_u1_marshal_VOID__STRING_STRING,
					   G_TYPE_NONE,
					   G_TYPE_STRING, G_TYPE_STRING,
					   G_TYPE_INVALID);
	dbus_g_object_register_marshaller (g_cclosure_marshal_VOID__VOID,
					   G_TYPE_NONE,
					   G_TYPE_INVALID);
	dbus_g_object_register_marshaller (g_cclosure_marshal_VOID__STRING,
					   G_TYPE_NONE,
					   G_TYPE_STRING,
					   G_TYPE_INVALID);

	/* Initialize syncdaemon */
	syncdaemon_init ();

	/* Create web viewer object */
	music_store->priv->scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (music_store->priv->scroll),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_widget_show (music_store->priv->scroll);

	music_store->priv->web_viewer = webkit_web_view_new ();

	new_user_agent = g_strdup_printf ("%s U1MusicStore/" VERSION ,
					  webkit_web_settings_get_user_agent (
						  webkit_web_view_get_settings (WEBKIT_WEB_VIEW (music_store->priv->web_viewer))));
	
	g_object_set (G_OBJECT (webkit_web_view_get_settings (WEBKIT_WEB_VIEW (music_store->priv->web_viewer))),
		      "user-agent", new_user_agent,
		      "enable-default-context-menu", g_getenv ("U1SHOWCONTEXTMENU") != NULL,
		      NULL);

	g_free(new_user_agent);

	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "load-finished",
			  G_CALLBACK (load_finished_cb), music_store);
	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "status-bar-text-changed",
			  G_CALLBACK (status_bar_text_changed_cb), music_store);
	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "load-error",
			  G_CALLBACK (load_error_cb), music_store);

	/* Show a temp view when loading the initial store page */
	music_store->priv->initial_web_viewer = webkit_web_view_new ();
	g_object_set (G_OBJECT (webkit_web_view_get_settings (WEBKIT_WEB_VIEW (music_store->priv->initial_web_viewer))),
		      "enable-default-context-menu", g_getenv ("U1SHOWCONTEXTMENU") != NULL,
		      NULL);
	webkit_web_view_load_string (WEBKIT_WEB_VIEW (music_store->priv->initial_web_viewer),
				     U1_INITIAL_PAGE,
				     "text/html", "utf-8", "file:///");
	gtk_widget_show (music_store->priv->initial_web_viewer);
	gtk_container_add (GTK_CONTAINER (music_store->priv->scroll), music_store->priv->initial_web_viewer);

	/* And load the real store in the background */
	g_idle_add ((GSourceFunc) load_real_store_cb, music_store);

	gtk_box_pack_start (GTK_BOX (music_store), music_store->priv->scroll, TRUE, TRUE, 3);
}

/**
 * u1_music_store_new:
 *
 * Create a new #U1MusicStore widget.
 *
 * Return value: the newly created widget.
 */
GtkWidget *
u1_music_store_new (void)
{
	U1MusicStore *music_store;

	music_store = g_object_new (U1_TYPE_MUSIC_STORE, NULL);

	return (GtkWidget *) music_store;
}

/**
 * u1_music_store_get_library_location:
 * @music_store: A #U1MusicStore object.
 *
 * Return the location where purchased music is stored on the user's
 * computer.
 *
 * Return value: Location of the music library.
 */
const gchar *
u1_music_store_get_library_location (U1MusicStore *music_store)
{
	static gchar *full_path = NULL;

	if (full_path == NULL) {
		full_path = g_build_filename (g_get_home_dir (),
					      U1_MUSIC_LIBRARY_LOCATION,
					      NULL);
	}

	return (const gchar *) full_path;
}
